from __future__ import annotations

from collections.abc import Callable
from collections.abc import Iterable as PyIterable
from enum import Enum as PyEnum
from io import BufferedIOBase, BufferedReader, BufferedWriter, TextIOWrapper
from types import FunctionType

from samarium.classes import (
    NULL,
    Array,
    Attrs,
    Enum,
    File,
    Function,
    Iterator,
    Mode,
    Null,
    Num,
    Number,
    Slice,
    String,
    Table,
    Zip,
)
from samarium.utils import get_type_name


class SliceRange:
    def __init__(self, slice_: Slice) -> None:
        self._slice = slice_

    @property
    def slice(self) -> slice:
        return self._slice.val

    @property
    def range(self) -> range:
        return self._slice.range


def to_python(obj: object) -> object:
    if isinstance(obj, (String, Number, Zip, File)):
        return obj.val
    if isinstance(obj, Null):
        return None
    if isinstance(obj, Array):
        return [to_python(i) for i in obj.val]
    if isinstance(obj, Table):
        return {to_python(k): to_python(v) for k, v in obj.val.items()}
    if isinstance(obj, Slice):
        return SliceRange(obj)
    if isinstance(obj, Enum):
        o = {k.removeprefix("sm_"): to_python(v) for k, v in obj.members.items()}
        return PyEnum(obj.name, o)
    if isinstance(obj, Iterator):
        return map(to_python, obj)
    msg = f"Conversion for type {type(obj).__name__!r} not found"
    raise TypeError(msg)


def to_samarium(obj: object) -> Attrs:
    if isinstance(obj, (int, bool, float)):
        return Num(obj)
    if isinstance(obj, str):
        return String(obj)
    if obj is None:
        return NULL
    if isinstance(obj, FunctionType):
        return Function(obj)
    if isinstance(obj, (list, tuple, set)):
        return Array([to_samarium(i) for i in obj])
    if isinstance(obj, dict):
        return Table({to_samarium(k): to_samarium(v) for k, v in obj.items()})
    if isinstance(obj, (range, slice)):
        return Slice(
            to_samarium(None if obj.start == 0 else obj.start),
            to_samarium(obj.stop),
            to_samarium(None if obj.step == 1 else obj.step),
        )
    if isinstance(obj, SliceRange):
        return obj._slice
    if isinstance(obj, (TextIOWrapper, BufferedWriter, BufferedReader)):
        return File(
            obj, Mode(obj.mode).name, obj.name, binary=isinstance(obj, BufferedIOBase)
        )
    if isinstance(obj, zip):
        return Zip(*obj)
    if isinstance(obj, type) and issubclass(obj, PyEnum):
        o = {f"sm_{k}": to_samarium(v) for k, v in obj.__members__.items()}
        return Enum(f"sm_{obj.__name__}", **o)
    if isinstance(obj, PyEnum):
        return to_samarium(obj.value)
    if isinstance(obj, PyIterable):
        return Iterator(obj)
    msg = f"Conversion for type {type(obj).__name__!r} not found"
    raise TypeError(msg)


def export(func: Callable) -> Callable[..., Attrs]:
    """Wraps a Python function to be used in Samarium"""

    if not isinstance(func, FunctionType):
        msg = f"cannot export a non-function type {get_type_name(func)!r}"
        raise TypeError(msg)

    def wrapper(*_args: Attrs) -> Attrs:
        return to_samarium(func(*map(to_python, _args)))

    f = Function(wrapper)
    f.__pyexported__ = True

    return f



from __future__ import annotations

from datetime import datetime, timezone
from re import compile
from time import sleep as _sleep
from time import time_ns
from types import GeneratorType
from typing import TYPE_CHECKING, Any, TypeVar

from samarium.classes import (
    MISSING,
    NULL,
    Array,
    Attrs,
    Null,
    Num,
    Number,
    Slice,
    String,
    correct_type,
)
from samarium.exceptions import (
    SamariumError,
    SamariumIOError,
    SamariumSyntaxError,
    SamariumTypeError,
)

if TYPE_CHECKING:
    from collections.abc import Callable

MISSING_ARGS_PATTERN = compile(
    r"\w+\(\) takes exactly one argument \(0 given\)"
    r"|\w+\(\) missing (\d+) required positional argument"
)

NULL_STRING = String()

T = TypeVar("T")

TOO_MANY_ARGS_PATTERN = compile(
    r"\w+\(\) takes (\d+) positional arguments? but (\d+) (?:was|were) given"
)


def dtnow() -> Array:
    utcnow = datetime.now(tz=timezone.utc)
    now = datetime.now().timetuple()
    utcnow_tt = utcnow.timetuple()
    tz = now[3] - utcnow_tt[3], now[4] - utcnow_tt[4]
    utcnow_tpl = utcnow_tt[:-3] + (utcnow.microsecond // 1000,) + tz
    return Array(map(Num, utcnow_tpl))


def mkslice(start: Any = None, stop: Any = MISSING, step: Any = None) -> Any:
    if stop is MISSING:
        if start is not None:
            return start
        return Slice(NULL)
    start = NULL if start is None else start
    stop = NULL if stop is None else stop
    step = NULL if step is None else step
    return Slice(start, stop, step)


def print_safe(*args: Attrs | Callable[..., Any] | bool | None) -> Attrs:
    typechecked_args = list(map(correct_type, args))
    return_args = typechecked_args.copy()
    strs = list(map(str, typechecked_args))
    types = list(map(type, typechecked_args))
    if tuple in types:
        msg = "missing brackets"
        raise SamariumSyntaxError(msg)
    if GeneratorType in types:
        msg = "invalid comprehension"
        raise SamariumSyntaxError(msg)
    print(*strs)
    if len(return_args) > 1:
        return Array(return_args)
    if not return_args or types[0] is Null:
        return NULL
    return return_args[0]  # type: ignore[return-value]


def readline(prompt: String = NULL_STRING) -> String:
    try:
        return String(input(prompt.val))
    except KeyboardInterrupt:
        msg = "^C"
    except EOFError:
        msg = "^D"
    raise SamariumIOError(msg)


def sleep(*args: Number) -> None:
    if not args:
        msg = "no argument provided for ,.,"
        raise SamariumTypeError(msg)
    if len(args) > 1:
        msg = ",., only takes one argument"
        raise SamariumTypeError(msg)
    if not isinstance((time := args[0]), Number):
        msg = ",., only accepts integers"
        raise SamariumTypeError(msg)
    _sleep(time.val / 1000)


def t(obj: T | None = None) -> T | None:
    return obj


def throw(obj: String | Array[Any] = NULL_STRING) -> None:
    note = ""
    if isinstance(obj, Array):
        if len(obj.val) != 2:
            msg = "array must be of form [error_msg, note]"
            raise SamariumTypeError(msg)
        error_msg, note = obj.val
        if not isinstance(error_msg, String):
            msg = "error message must be a string"
            raise SamariumTypeError(msg)
        if not isinstance(note, String):
            msg = "error note must be a string"
            raise SamariumTypeError(msg)
        note = f"\n&1[Note] {note.val}"
    elif isinstance(obj, String):
        error_msg = obj
    else:
        msg = "throw argument must be a string or a 2-element array"
        raise SamariumTypeError(msg)
    raise SamariumError(error_msg.val + note)


def timestamp() -> Number:
    return Num(time_ns() // 1_000_000)



from __future__ import annotations

import re
from collections.abc import Callable, Iterable, Mapping
from collections.abc import Iterator as PyIterator
from contextlib import suppress
from functools import lru_cache
from inspect import signature
from random import choice, randrange, uniform
from types import GeneratorType, MethodType
from typing import Any, Generic, TypeVar, cast

from samarium.exceptions import (
    NotDefinedError,
    SamariumSyntaxError,
    SamariumTypeError,
    SamariumValueError,
)
from samarium.utils import (
    ClassProperty,
    Singleton,
    get_name,
    get_type_name,
    parse_number,
    smformat,
)

T = TypeVar("T")
KT = TypeVar("KT")
VT = TypeVar("VT")

MISSING_POSARG = re.compile(r"missing (\d+) required positional argument")


def functype_repr(obj: Any) -> str:
    return (
        get_name(obj)
        if isinstance(obj, type) and issubclass(obj, UserAttrs)
        else repr(obj)
    )


def guard(operator: str, *, default: int | None = None) -> Callable[..., Any]:
    def decorator(function: Callable[..., Any]) -> Callable[..., Any]:
        def wrapper(self: Any, other: Any) -> Any:
            cls = type(self)
            use_default = default is not None and other.val is None
            if isinstance(other, cls):
                return function(self, other)
            if use_default:
                return function(self, Num(default))
            msg = f"{cls.__name__} {operator} {type(other).__name__}"
            raise NotDefinedError(msg)

        return wrapper

    return decorator


def throw_missing(*_: Any) -> None:
    msg = "missing default parameter value(s)"
    raise SamariumTypeError(msg)


class Missing:
    @classmethod
    def type(cls) -> None:
        throw_missing()

    @classmethod
    def parent(cls) -> None:
        throw_missing()

    @classmethod
    def id(cls) -> None:
        throw_missing()


METHODS = (
    "getattr add str sub mul truediv pow mod and or xor neg pos invert getitem "
    "setitem eq ne ge gt le lt contains hash call iter matmul"
)

for m in METHODS.split():
    setattr(Missing, f"__{m}__", throw_missing)

MISSING = Missing()
NEXT = object()


class CompositionMeta(type):
    def __and__(cls, other: Any) -> Function:
        if isinstance(other, type):
            other = Type(other)
        if isinstance(other, Type):
            other = other.as_function()
        if not isinstance(other, Function):
            msg = f"can't use function composition with {get_type_name(other)}"
            raise SamariumTypeError(msg)
        return Type(cls).as_function() & other


class Attrs(metaclass=CompositionMeta):
    def __getattribute__(self, name: str) -> Any:
        attr = object.__getattribute__(self, name)
        if isinstance(attr, Function) and "self" in signature(attr.func).parameters:
            return Function(attr.func, self)
        return attr

    @classmethod
    def id(cls) -> String:
        return String(f"{id(cls):x}")

    def __repr__(self) -> str:
        return str(self)

    @classmethod
    def type(cls) -> Type:
        return Type(cls)

    parent = type


class UserAttrs(Attrs):
    def __entry__(self, *args: object) -> None: ...

    def __init__(self, *args: object) -> None:
        self.__entry__(*args)

    def __string__(self) -> String:
        return String(f"<{get_type_name(self)}@{self.id()}>")

    __string__.param_count = 1

    def __str__(self) -> str:
        string = self.__string__
        if param_count(string) != 1:
            msg = f"{get_type_name(self)}! should only take one argument"
            raise SamariumTypeError(msg)
        if isinstance(v := string().val, str):
            return v
        msg = f"{get_type_name(self)}! returned a non-string"
        raise SamariumTypeError(msg)

    def __bit__(self) -> Number:
        msg = f"? {get_type_name(self)}"
        raise NotDefinedError(msg)

    __bit__.param_count = 1

    def __bool__(self) -> bool:
        bit = self.__bit__
        if param_count(bit) != 1:
            msg = f"? {get_type_name(self)} should only take one argument"
            raise SamariumTypeError(msg)
        if (v := bit().val) in (0, 1):
            return bool(v)
        msg = f"? {get_type_name(self)} returned a non-bit"
        raise SamariumTypeError(msg)

    def __hash__(self) -> int:
        return cast(int, self.hash().val)

    def __special__(self) -> object:
        msg = f"{get_type_name(self)}$"
        raise NotDefinedError(msg)

    __special__.param_count = 1

    def special(self) -> Any:
        special = self.__special__
        if param_count(special) != 1:
            msg = f"{get_type_name(self)}$ should only take one argument"
            raise SamariumTypeError(msg)
        return special()

    def __hsh__(self) -> Number:
        msg = f"{get_type_name(self)}##"
        raise NotDefinedError(msg)

    __hsh__.param_count = 1

    def hash(self) -> Number:
        hsh = self.__hsh__
        if param_count(hsh) != 1:
            msg = f"{get_type_name(self)}## should only take one argument"
            raise SamariumTypeError(msg)
        if isinstance(v := hsh(), Number):
            return v
        msg = f"{get_type_name(self)}## returned a non-integer"
        raise SamariumTypeError(msg)

    def __cast__(self) -> object:
        msg = f"{get_type_name(self)}%"
        raise NotDefinedError(msg)

    __cast__.param_count = 1

    def cast(self) -> object:
        cast = self.__cast__
        if param_count(cast) != 1:
            msg = f"{get_type_name(self)}% should only take one argument"
            raise SamariumTypeError(msg)
        return cast()

    def __random__(self) -> Number:
        msg = f"{get_type_name(self)}??"
        raise NotDefinedError(msg)

    __random__.param_count = 1

    def random(self) -> object:
        random = self.__random__
        if param_count(random) != 1:
            msg = f"{get_type_name(self)}?? should only take one argument"
            raise SamariumTypeError(msg)
        return random()

    @ClassProperty
    def param_count(self) -> int:
        return len(signature(self.__init__).parameters)  # type: ignore[misc]

    @classmethod
    def parent(cls) -> Array | Type:  # type: ignore [override]
        parents = cls.__bases__
        if len(parents) == 1:
            parent = parents[0]
            if parent is object:
                return Type(cls)
            if parent is UserAttrs:
                return Type(Type)
            return Type(parent)
        return Array(map(Type, parents))


class Dataclass(UserAttrs):
    _name: str
    _field_names: list[str]
    _fields: list[str]

    def __init_subclass__(cls, /, fields: list[str]) -> None:
        super().__init_subclass__()
        cls._name = cls.__name__.removeprefix("sm_")
        cls._field_names = fields
        cls._fields = [f"sm_{f}" for f in fields]

    def __init__(self, *args: Attrs) -> None:
        argc = len(args)
        fieldc = len(self._fields)
        if argc < fieldc:
            msg = f"missing argument(s): {', '.join(self._field_names[argc:])}"
            raise SamariumTypeError(msg)
        if argc > fieldc:
            msg = f"too many arguments ({argc}/{fieldc})"
            raise SamariumTypeError(msg)
        for field, arg in zip(self._fields, args):
            setattr(self, field, arg)

    @property
    def vals(self) -> tuple[Attrs, ...]:
        return tuple(getattr(self, i) for i in self._fields)

    def __str__(self) -> str:
        return f"{self._name}({', '.join(map(repr, self.vals))})"

    def __eq__(self, other: object) -> bool:
        if isinstance(other, type(self)):
            return self.vals == other.vals
        return False

    def __gt__(self, other: object) -> bool:
        if isinstance(other, type(self)):
            return self.vals > other.vals
        return False

    def __ne__(self, other: object) -> bool:
        return not (self == other)

    def __ge__(self, other: object) -> bool:
        return self == other or self > other

    def __lt__(self, other: object) -> bool:
        return not (self >= other)

    def __le__(self, other: object) -> bool:
        return not (self > other)

    def special(self) -> Dataclass:
        return type(self)(*self.vals)


class Type(Attrs):
    __slots__ = ("val",)

    def __init__(self, type_: type) -> None:
        self.val = type_

    def __eq__(self, other: object) -> Number:
        if isinstance(other, Type):
            return Num(self.val == other.val)
        return Num(self.val == other)

    def __ne__(self, other: object) -> Number:
        if isinstance(other, Type):
            return Num(self.val != other.val)
        return Num(self.val != other)

    def as_function(self) -> Function:
        # "redundant" lambda on purpose to make the obj have a __code__ attr
        return Function(lambda x: self(x))

    def __and__(self, other: object) -> Function:
        if isinstance(other, type):
            other = Type(other)
        if isinstance(other, Type):
            other = other.as_function()
        if not isinstance(other, Function):
            msg = f"Type ++ {get_type_name(other)}"
            raise SamariumTypeError(msg)
        return self.as_function() & other

    def __str__(self) -> str:
        return get_name(self.val)

    __repr__ = __str__

    def __bool__(self) -> bool:
        return True

    def __hash__(self) -> int:
        return hash(self.val)

    def __call__(self, *args: object) -> object:
        if self.val in (Function, Module, Type, Dataclass):
            msg = f"cannot instantiate a {self.val.__name__.lower()}"
            raise SamariumTypeError(msg)
        return self.val(*args)


class Module(Attrs):
    __slots__ = ("name", "objects")

    def __init__(self, name: str, objects: Mapping[str, object]) -> None:
        self.name = name
        self.objects = objects

    def __bool__(self) -> bool:
        return True

    def __str__(self) -> str:
        return f"Module[{self.name}]"

    __repr__ = __str__

    def __getattr__(self, key: str) -> object:
        return self.objects[key]


class Number(Attrs):
    __slots__ = ("is_int", "val")

    def __init__(self, v: Any = None) -> None:
        self.val: int | float
        t = type(v)
        if t in (int, bool):
            self.is_int = True
            self.val = int(v)
        elif t is float:
            self.is_int = is_int = v.is_integer()
            self.val = int(v) if is_int else v
        elif t is Number:
            self.val = v.val
            self.is_int = v.is_int
        elif t is String:
            val, is_int = parse_number(v.val) if v else (0, True)
            self.val = int(val) if is_int else val
            self.is_int = is_int
        elif v in (None, NULL):
            self.val = 0
            self.is_int = True
        else:
            msg = f"cannot cast {get_name(t)} to Number"
            raise SamariumTypeError(msg)

    def __bool__(self) -> bool:
        return self.val != 0.0

    def __str__(self) -> str:
        return str(self.val)

    __repr__ = __str__

    @guard("+", default=1)
    def __add__(self, other: Any) -> Number:
        return Num(self.val + other.val)

    @guard("-", default=1)
    def __sub__(self, other: Any) -> Number:
        return Num(self.val - other.val)

    def __mul__(self, other: Any) -> String | Number:
        if isinstance(other, String):
            return other * self
        if other is NULL:
            other = Num(2)
        return Num(self.val * other.val)

    @guard("--", default=2)
    def __truediv__(self, other: Any) -> Number:
        return Num(self.val / other.val)

    @guard("+++", default=2)
    def __pow__(self, other: Any) -> Number:
        return Num(self.val**other.val)

    @guard("---", default=2)
    def __mod__(self, other: Any) -> Number:
        return Num(self.val % other.val)

    @guard("&")
    def __and__(self, other: Any) -> Number:
        if self.is_int and other.is_int:
            return Num(self.val & other.val)
        msg = "cannot use & with non-integer numbers"
        raise SamariumValueError(msg)

    @guard("|")
    def __or__(self, other: Any) -> Number:
        if self.is_int and other.is_int:
            return Num(self.val | other.val)
        msg = "cannot use | with non-integer numbers"
        raise SamariumValueError(msg)

    @guard("^")
    def __xor__(self, other: Any) -> Number:
        if self.is_int and other.is_int:
            return Num(self.val ^ other.val)
        msg = "cannot use ^ with non-integer numbers"
        raise SamariumValueError(msg)

    def __invert__(self) -> Number:
        if self.is_int:
            return Num(~cast(int, self.val))
        msg = "cannot invert a non-integer number"
        raise SamariumValueError(msg)

    def __pos__(self) -> Number:
        return self

    def __neg__(self) -> Number:
        return Num(-self.val)

    def __eq__(self, other: object) -> Number:
        if isinstance(other, Number):
            return Num(self.val == other.val)
        return Num(0.0)

    def __ne__(self, other: object) -> Number:
        if isinstance(other, Number):
            return Num(self.val != other.val)
        return Num(1.0)

    @guard(">", default=0)
    def __gt__(self, other: Any) -> Number:
        return Num(self.val > other.val)

    @guard(">:", default=0)
    def __ge__(self, other: Any) -> Number:
        return Num(self.val >= other.val)

    @guard("<", default=0)
    def __lt__(self, other: Any) -> Number:
        return Num(self.val < other.val)

    @guard("<:", default=0)
    def __le__(self, other: Any) -> Number:
        return Num(self.val <= other.val)

    def __hash__(self) -> int:
        return cast(int, self.hash().val)

    def cast(self) -> String:
        if self.is_int:
            return String(to_chr(cast(int, self.val)))
        msg = "cannot cast a non-integer number"
        raise SamariumValueError(msg)

    def hash(self) -> Number:
        return Num(hash(self.val))

    def random(self) -> Number:
        if not self:
            return self
        if self.is_int:
            r = randrange(v_ := cast(int, self.val))
            return Num(r if v_ > 0 else ~r)
        u = uniform(0, v := self.val)
        if v > 0.0:
            return Num(u)
        return Num(-u - 1)

    def special(self) -> Number:
        return Number(int(self.val))


Num = lru_cache(1024)(Number)


class String(Attrs):
    __slots__ = ("val",)

    def __init__(self, value: object = "") -> None:
        self.val = str(value)

    def __contains__(self, element: String) -> bool:
        return element.val in self.val

    def __iter__(self) -> PyIterator[String]:
        yield from map(String, self.val)

    def __bool__(self) -> bool:
        return self.val != ""

    def __str__(self) -> str:
        return self.val

    def __repr__(self) -> str:
        return f'"{repr(self.val)[1:-1]}"'

    def __neg__(self) -> String:
        return String(self.val[::-1])

    def __add__(self, other: object) -> String:
        if isinstance(other, String):
            return String(self.val + other.val)
        if other is NULL:
            other = Num(1)
        if isinstance(other, Number):
            if other.is_int:
                return String(
                    "".join(
                        chr((ord(i) + cast(int, other.val)) % 0x10FFFF)
                        for i in self.val
                    )
                )
            msg = "cannot shift using non-integers"
            raise SamariumTypeError(msg)
        msg = f"String + {get_type_name(other)}"
        raise SamariumTypeError(msg)

    def __sub__(self, other: object) -> String:
        if isinstance(other, String):
            return String(self.val.replace(other.val, "", 1))
        if isinstance(other, Number):
            return self + (-other)
        msg = f"String - {get_type_name(other)}"
        raise SamariumTypeError(msg)

    @guard("--")
    def __truediv__(self, other: Any) -> String:
        return String(self.val.replace(other.val, ""))

    def __mul__(self, other: object) -> String:
        if other is NULL:
            other = Num(2)
        if isinstance(other, Number):
            if other.val < 0:
                msg = (
                    "String ++ -Number is ambiguous, use -(String ++ Number)"
                    " or (-String) ++ Number instead"
                )
                raise SamariumTypeError(msg)
            i, d = int(other.val // 1), other.val % 1
            return String(self.val * i + self.val[: round(d * len(self.val))])
        msg = f"String ++ {get_type_name(other)}"
        raise NotDefinedError(msg)

    def __mod__(self, other: object) -> String:
        if isinstance(other, (String, Array, Table)):
            return String(smformat(self.val, other.val))
        msg = f"String --- {get_type_name(other)}"
        raise NotDefinedError(msg)

    def __eq__(self, other: object) -> Number:
        return Num(self.val == other.val)

    def __ne__(self, other: object) -> Number:
        return Num(self.val != other.val)

    @guard(">")
    def __gt__(self, other: Any) -> Number:
        return Num(self.val > other.val)

    @guard(">:")
    def __ge__(self, other: Any) -> Number:
        return Num(self.val >= other.val)

    @guard("<")
    def __lt__(self, other: Any) -> Number:
        return Num(self.val < other.val)

    @guard("<:")
    def __le__(self, other: Any) -> Number:
        return Num(self.val <= other.val)

    def __getitem__(self, index: Number | Slice) -> String:
        if isinstance(index, Number):
            if not is_valid_index(self, index):
                msg = f"invalid index: {index}"
                raise SamariumTypeError(msg)
            return String(self.val[cast(int, index.val)])
        if isinstance(index, Slice):
            return String(self.val[index.val])
        msg = f"invalid index: {index}"
        raise SamariumTypeError(msg)

    def __setitem__(self, index: Number | Slice, value: String) -> None:
        if not isinstance(index, (Number, Slice)):
            msg = f"invalid index: {index}"
            raise SamariumTypeError(msg)
        i: int | slice
        if isinstance(index, Number):
            if not is_valid_index(self, index):
                msg = f"invalid index: {index}"
                raise SamariumTypeError(msg)
            i = cast(int, index.val)
        else:
            i = index.val
        string = [*self.val]
        string[i] = value.val
        self.val = "".join(string)

    def __hash__(self) -> int:
        return cast(int, self.hash().val)

    def __matmul__(self, other: object) -> Zip:
        return Zip(self, other)

    def cast(self) -> Array | Number:
        if len(self.val) == 1:
            return Num(ord(self.val))
        return Array(Num(ord(i)) for i in self.val)

    def hash(self) -> Number:
        return Num(hash(self.val))

    def random(self) -> String:
        return String(choice(self.val))

    def special(self) -> Number:
        return Num(len(self.val))


class Array(Generic[T], Attrs):
    __slots__ = ("val",)

    def __init__(self, value: object = None) -> None:
        self.val: list[Any]
        if value is None:
            self.val = []
        elif isinstance(value, Array):
            self.val = value.val.copy()
        elif isinstance(value, String):
            self.val = list(map(String, value.val))
        elif isinstance(value, Table):
            self.val = list(map(Array, value.val.items()))
        elif isinstance(value, Slice):
            if value.stop is NULL:
                msg = "cannot convert an infinite Slice to an Array"
                raise SamariumValueError(msg)
            self.val = list(value)
        elif isinstance(value, Iterable):
            self.val = [*value]
        else:
            msg = f"cannot cast {get_type_name(value)} to an Array"
            raise SamariumTypeError(msg)

    def __str__(self) -> str:
        return ", ".join(map(functype_repr, self.val)).join("[]")

    __repr__ = __str__

    def __bool__(self) -> bool:
        return self.val != []

    def __iter__(self) -> PyIterator[T]:
        yield from self.val

    def __contains__(self, element: T) -> bool:
        return element in self.val

    def __eq__(self, other: object) -> Number:
        if isinstance(other, Array):
            return Num(self.val == other.val)
        return Num(0)

    def __ne__(self, other: object) -> Number:
        if isinstance(other, Array):
            return Num(self.val != other.val)
        return Num(1)

    @guard(">")
    def __gt__(self, other: Any) -> Number:
        return Num(self.val > other.val)

    @guard(">:")
    def __ge__(self, other: Any) -> Number:
        return Num(self.val >= other.val)

    @guard("<")
    def __lt__(self, other: Any) -> Number:
        return Num(self.val < other.val)

    @guard("<:")
    def __le__(self, other: Any) -> Number:
        return Num(self.val <= other.val)

    def __getitem__(self, index: object) -> T | Array[T]:
        if isinstance(index, Number):
            if not is_valid_index(self, index):
                msg = f"invalid index: {index}"
                raise SamariumValueError(msg)
            return self.val[cast(int, index.val)]
        if isinstance(index, Slice):
            return Array(self.val[index.val])
        msg = f"invalid index: {index}"
        raise SamariumTypeError(msg)

    def __setitem__(self, index: object, value: T | Array[T]) -> None:
        if not isinstance(index, (Number, Slice)):
            msg = f"invalid index: {index}"
            raise SamariumTypeError(msg)
        if isinstance(index, Number):
            if not is_valid_index(self, index):
                msg = f"invalid index: {index}"
                raise SamariumValueError(msg)
            self.val[cast(int, index.val)] = value
        else:
            self.val[index.val] = cast(Array[T], value)

    @guard("+")
    def __add__(self, other: Any) -> Array[T]:
        return Array(self.val + other.val)

    def __sub__(self, other: object) -> Array[T]:
        new_array = self.val.copy()
        if isinstance(other, Array):
            for i in other:
                try:
                    new_array.remove(i)
                except ValueError:
                    msg = f"{i!r} not in array"
                    raise SamariumValueError(msg) from None
        elif isinstance(other, Number):
            if not other.is_int:
                msg = f"invalid index: {other}"
                raise SamariumValueError(msg)
            new_array.pop(cast(int, other.val))
        elif other is NULL:
            return self.val.pop()
        else:
            msg = f"Array - {get_type_name(other)}"
            raise NotDefinedError(msg)
        return Array(new_array)

    def __neg__(self) -> Array[T]:
        new_array = []
        for i in self.val:
            if i not in new_array:
                new_array.append(i)
        return Array(new_array)

    def __mod__(self, other: object) -> Array[T]:
        indexes = [i for i, v in enumerate(self.val) if v == other]
        new_array = self.val.copy()
        for i in indexes[1:][::-1]:
            new_array.pop(i)
        return Array(new_array)

    @guard("--")
    def __truediv__(self, other: Any) -> Array[T]:
        new_array = self.val.copy()
        for i in other.val:
            with suppress(ValueError):
                new_array.remove(i)
        return Array(new_array)

    @guard("|")
    def __or__(self, other: Any) -> Array[T]:
        new_array = self.val.copy()
        for i in other.val:
            if i not in new_array:
                new_array.append(i)
        return Array(new_array)

    @guard("&")
    def __and__(self, other: Any) -> Array[T]:
        return Array([i for i in other.val if i in self.val])

    @guard("^")
    def __xor__(self, other: Any) -> Array[T]:
        return (self | other) / (self & other)

    def __mul__(self, other: object) -> Array[T]:
        if isinstance(other, Number):
            i, d = int(other.val // 1), other.val % 1
            return Array(self.val * i + self.val[: round(d * len(self.val))])
        msg = f"Array ++ {get_type_name(other)}"
        raise NotDefinedError(msg)

    def __matmul__(self, other: object) -> Zip:
        return Zip(self, other)

    def cast(self) -> String:
        s = ""
        for i in self.val:
            if isinstance(i, Number) and i.is_int:
                s += to_chr(cast(int, i.val))
            else:
                msg = "array contains non-integers"
                raise SamariumTypeError(msg)
        return String(s)

    def random(self) -> T:
        if not self.val:
            msg = "array is empty"
            raise SamariumValueError(msg)
        return choice(self.val)

    def special(self) -> Number:
        return Num(len(self.val))


class Table(Generic[KT, VT], Attrs):
    __slots__ = ("val",)

    def __init__(self, value: object = None) -> None:
        if value is None:
            self.val: dict[KT, VT] = {}
        elif isinstance(value, Table):
            self.val = value.val.copy()
        elif isinstance(value, dict):
            self.val = value
        elif isinstance(value, Array):
            arr = value.val
            if all(isinstance(i, (String, Array)) and len(i.val) == 2 for i in arr):
                table = {}
                for e in arr:
                    if isinstance(e, String):
                        k, v = map(String, e.val)
                        table[k] = v
                    else:
                        k, v = e.val
                        table[k] = v
                self.val = cast(dict[KT, VT], table)
            else:
                msg = "not all elements of the array are of length 2"
                raise SamariumValueError(msg)
        else:
            msg = f"cannot cast {get_type_name(value)} to a Table"
            raise SamariumTypeError(msg)

    def __str__(self) -> str:
        return ", ".join(
            "{} -> {}".format(*map(functype_repr, i)) for i in self.val.items()
        ).join(("{{", "}}"))

    def __bool__(self) -> bool:
        return self.val != {}

    def __getitem__(self, key: KT) -> VT:
        try:
            return self.val[key]
        except KeyError:
            msg = f"key not found: {key}"
            raise SamariumValueError(msg) from None

    def __setitem__(self, key: KT, value: VT) -> None:
        self.val[key] = value

    def __iter__(self) -> PyIterator[KT]:
        yield from self.val.keys()

    def __contains__(self, element: object) -> bool:
        return element in self.val

    def __eq__(self, other: object) -> Number:
        if isinstance(other, Table):
            return Num(self.val == other.val)
        return Num(0)

    def __ne__(self, other: object) -> Number:
        if isinstance(other, Table):
            return Num(self.val != other.val)
        return Num(1)

    @guard("+")
    def __add__(self, other: Any) -> Table[KT, VT]:
        return Table(self.val | other.val)

    def __invert__(self) -> Table[VT, KT]:
        return Table({v: k for k, v in self.val.items()})

    def __sub__(self, other: KT) -> Table[KT, VT]:
        c = self.val.copy()
        try:
            del c[other]
        except KeyError:
            msg = f"key not found: {other}"
            raise SamariumValueError(msg) from None
        return Table(c)

    def __matmul__(self, other: object) -> Zip:
        return Zip(self, other)

    def random(self) -> object:
        if not self.val:
            msg = "table is empty"
            raise SamariumValueError(msg)
        return choice(list(self.val.keys()))

    def special(self) -> Array[VT]:
        return Array(self.val.values())


class Null(Singleton, Attrs):
    __slots__ = ("val",)

    def __init__(self) -> None:
        self.val = None

    def __str__(self) -> str:
        return "null"

    __repr__ = __str__

    def __bool__(self) -> bool:
        return False

    def __eq__(self, other: object) -> bool:
        return self is other

    def __ne__(self, other: object) -> bool:
        return self is not other

    def __hash__(self) -> int:
        return cast(int, self.hash().val)

    def hash(self) -> Number:
        return Num(hash(self.val))

    def __invert__(self) -> Number:
        return Num(-1)

    def __floordiv__(self, other: object) -> Number:
        return Num(not other)


I64_MAX = 9223372036854775807
NULL = Null()


class Slice(Attrs):
    __slots__ = ("start", "stop", "step", "tup", "range", "val")

    def __init__(self, start: Any, stop: Any = NULL, step: Any = NULL) -> None:
        if step.val == 0:
            msg = "step cannot be zero"
            raise SamariumValueError(msg)
        self.start = start
        self.stop = stop
        self.step = step
        self.tup = start.val, stop.val, step.val
        try:
            self.range = range(
                self.tup[0] or 0,
                I64_MAX if self.tup[1] is None else self.tup[1],
                self.tup[2] or 1,
            )
        except TypeError:
            msg = "slice values have to be integers"
            raise SamariumTypeError(msg) from None
        self.val = slice(*self.tup)

    def __bool__(self) -> bool:
        return bool(self.range)

    def __iter__(self) -> PyIterator[Number]:
        yield from map(Num, self.range)

    def __contains__(self, value: Number) -> Number:
        return Num(value.val in self.range)

    def __getitem__(self, index: Number) -> Number:
        if not isinstance(index, Number):
            msg = f"invalid index: {index}"
            raise SamariumTypeError(msg)
        if is_valid_index(self, index):
            return Num(self.range[cast(int, index.val)])
        msg = f"invalid index: {index}"
        raise SamariumValueError(msg)

    def __str__(self) -> str:
        start, stop, step = self.tup
        if self.is_empty():
            return "<<>>"
        string = ""
        if start is not None:
            string += repr(start)
            if stop is step is None:
                string += ".."
        if stop is not None:
            string += f"..{stop!r}"
        if step is not None:
            if stop is None:
                string += ".."
            string += f"..{step!r}"
        return f"<<{string}>>"

    def __matmul__(self, other: object) -> Zip:
        return Zip(self, other)

    __repr__ = __str__

    def __eq__(self, other: object) -> Number:
        if isinstance(other, Slice):
            return Num(self.tup == other.tup)
        return Num(0)

    def __ne__(self, other: object) -> Number:
        if isinstance(other, Slice):
            return Num(self.tup != other.tup)
        return Num(1)

    def __hash__(self) -> int:
        return cast(int, self.hash().val)

    def hash(self) -> Number:
        return Num(hash(self.tup))

    def random(self) -> Number:
        return Num(choice(self.range))

    def special(self) -> Number:
        return Num(len(self.range))

    def is_empty(self) -> bool:
        return self.start is self.stop is self.step is NULL


class Zip(Attrs):
    __slots__ = ("iters", "val")

    def __init__(self, *values: Any) -> None:
        if not all(isinstance(i, Iterable) for i in values):
            msg = "cannot zip a non-iterable"
            raise SamariumTypeError(msg)
        self.iters = values
        self.val = zip(*values)

    def __hash__(self) -> int:
        return hash(self.val)

    def __bool__(self) -> bool:
        return True

    def __iter__(self) -> PyIterator[Array]:
        yield from map(Array, self.val)

    def __matmul__(self, other: object) -> Zip:
        return Zip(*self.iters, other)

    def __str__(self) -> str:
        return f"<Zip@{id(self):x}>"

    def special(self) -> Number:
        return Num(len(self.iters))


class Iterator(Generic[T], Attrs):
    __slots__ = ("val", "length")

    length: Number | Null

    def __init__(self, value: Any) -> None:
        if not isinstance(value, Iterable):
            msg = "cannot create an Iterator from a non-iterable"
            raise SamariumTypeError(msg)
        self.val = iter(value)
        try:
            self.length = Num(len(value.val))
        except (TypeError, AttributeError):
            self.length = NULL

    def __bool__(self) -> bool:
        return True

    def __iter__(self) -> PyIterator[T]:
        yield from self.val

    def __next__(self) -> T:
        return next(self.val)

    def __str__(self) -> str:
        return f"<Iterator@{id(self):x}>"

    def cast(self) -> Number | Null:
        return self.length

    def special(self) -> T:
        return next(self)


class Enum(Attrs):
    __slots__ = ("name", "members")

    def __bool__(self) -> bool:
        return bool(self.members)

    def __init__(self, name: str, **members: object) -> None:
        self.name = name.removeprefix("sm_")
        self.members: dict[str, object] = {}

        i: int | float = 0
        for k, v in members.items():
            if v is NEXT:
                self.members[k] = Num(i)
                i += 1
            else:
                self.members[k] = v
                if isinstance(v, Number):
                    i = v.val + 1

    def __str__(self) -> str:
        return f"Enum({self.name})"

    def __getattr__(self, name: str) -> object:
        try:
            return self.members[name]
        except KeyError:
            msg = f"'{self.name}''{name}'"
            raise AttributeError(msg) from None

    def __setattr__(self, name: str, value: object) -> None:
        if name.startswith("sm_"):
            msg = "enum members cannot be modified"
            raise SamariumTypeError(msg)
        object.__setattr__(self, name, value)

    def cast(self) -> Table:
        return Table(
            {
                String(k.removeprefix("__").removeprefix("sm_")): v
                for k, v in self.members.items()
            }
        )


class Function(Attrs):
    __pyexported__ = False

    def __init__(
        self,
        func: Callable[..., Any],
        inst: Any | None = None,
    ) -> None:
        code = func.__code__
        flags = code.co_flags
        self.varargs = bool(flags & 4)
        func.__code__ = func.__code__.replace(
            co_flags=flags & ~4, co_argcount=code.co_argcount + self.varargs
        )
        self.func = func
        self.param_count = len(signature(func).parameters) - self.varargs
        self.inst = inst

    def __str__(self) -> str:
        return get_name(self.func)

    def __call__(self, *args: Any) -> Any:
        for arg in args:
            check_type(arg)
        supplied = len(args)
        posargs, args = args[: self.param_count], args[self.param_count :]
        if self.inst is not None:
            posargs = (self.inst, *posargs)
        if not self.varargs and args:
            msg = f"too many arguments ({supplied}/{self.param_count})"
            raise SamariumTypeError(msg)
        try:
            out = correct_type(
                self.func(*posargs, Array(args))
                if self.varargs
                else self.func(*posargs)
            )
        except TypeError as e:
            errmsg = str(e)
            if m := MISSING_POSARG.search(errmsg):
                msg = f"not enough arguments ({supplied}/{supplied + int(m.group(1))})"
                raise SamariumTypeError(msg) from None
            if "positional argument: 'self'" in errmsg:
                msg = "missing instance"
                raise SamariumTypeError(msg) from None
            raise
        return out

    def __and__(self, other: Any) -> Function:
        if isinstance(other, type):
            other = Type(other)
        if isinstance(other, Type):
            other = other.as_function()
        if not isinstance(other, Function):
            msg = f"Function & {get_type_name(other)}"
            raise SamariumTypeError(msg)

        def f(*args: Attrs) -> Attrs:
            return self(other(*args))

        f.__name__ = f"({self} & {other})"
        return Function(f)

    def __hash__(self) -> int:
        return cast(int, self.hash().val)

    def hash(self) -> Number:
        return Num(hash(self.func))

    def special(self) -> Number:
        return Num(self.param_count + self.varargs)


def check_type(obj: Any) -> None:
    if isinstance(obj, property):
        msg = "cannot use a special method on a type"
        raise SamariumTypeError(msg)
    if isinstance(obj, (tuple, GeneratorType)):
        msg = "invalid syntax"
        raise SamariumSyntaxError(msg)


def correct_type(obj: T, *objs: T) -> T | Attrs:
    if objs:
        return Array(map(correct_type, (obj, *objs)))  # type: ignore[arg-type]
    if isinstance(
        obj, (Null, Number, String, Slice, Enum, Type, Module, Zip, Function)
    ):
        return obj
    if obj is None:
        return NULL
    if isinstance(obj, bool):
        return Num(obj)
    if isinstance(obj, GeneratorType):
        return Iterator(obj)
    if isinstance(obj, (list, tuple, Array)):
        return Array(map(correct_type, obj))
    if isinstance(obj, Table):
        return Table({correct_type(k): correct_type(v) for k, v in obj.val.items()})
    check_type(obj)
    return obj


def is_valid_index(obj: Attrs, index: Number) -> bool:
    len_ = len(obj.val)
    return -len_ <= index.val < len_ and index.is_int


def param_count(func: Callable) -> int:
    with suppress(AttributeError):
        return func.param_count
    return len(signature(func).parameters) + isinstance(func, MethodType)


def to_chr(code: int) -> str:
    if 0 <= code < 0x110000:
        return chr(code)
    msg = "invalid Unicode code point"
    raise SamariumValueError(msg)



# ruff: noqa: F401
from __future__ import annotations

import ast
import importlib.machinery
import importlib.util
import sys
from pathlib import Path
from typing import TYPE_CHECKING

from samarium import exceptions as exc
from samarium.builtins import (
    dtnow,
    mkslice,
    print_safe,
    readline,
    sleep,
    t,
    throw,
    timestamp,
)
from samarium.classes import (
    MISSING,
    NEXT,
    NULL,
    Array,
    Dataclass,
    Enum,
    FileManager,
    Function,
    Mode,
    Module,
    Null,
    Num,
    Number,
    String,
    Table,
    UserAttrs,
    correct_type,
)
from samarium.exceptions import DAHLIA
from samarium.imports import merge_objects, parse_string, resolve_path
from samarium.runtime import Runtime
from samarium.tokenizer import tokenize
from samarium.transpiler import Registry, Transpiler
from samarium.utils import sysexit

if TYPE_CHECKING:
    from samarium.classes import Attrs


def import_to_scope(data: str, reg: Registry, source: str) -> None:
    modules = parse_string(data)
    for mod in modules:
        if mod.name == "samarium":
            raise exc.SamariumRecursionError
        path = resolve_path(mod.name, source)
        mod_path = (path / f"{mod.name}.sm").resolve()
        if mod_path.exists():
            imported = run(mod_path.read_text(), Registry({}), mod_path)
        else:
            spec = importlib.util.spec_from_file_location(
                mod.name, str(path / f"{mod.name}.py")
            )
            if spec is None:
                msg = "couldn't load spec"
                raise ValueError(msg)
            module = importlib.util.module_from_spec(spec)
            sys.modules[mod.name] = module
            if spec.loader is None:
                msg = "ModuleSpec.loader is None"
                raise ValueError(msg)
            spec.loader.exec_module(module)
            registry = {
                f"sm_{k}": v
                for k, v in vars(module).items()
                if getattr(v, "__pyexported__", False)
            }
            imported = Registry(registry)

        reg.vars.update(merge_objects(reg, imported, mod))


def import_inline(data: str, source: str) -> Attrs:
    reg = Registry({})
    import_to_scope(data, reg, source)
    return reg.vars.popitem()[1]


def run(
    code: str,
    reg: Registry,
    source: Path | str,
    *,
    debug: bool = False,
    load_template: bool = True,
    repl: bool = False,
) -> Registry:
    runtime_state = Runtime.repl
    Runtime.repl = repl
    code = Transpiler(tokenize(code), reg).transpile().output
    if load_template:
        code = (
            (Path(__file__).resolve().parent / "template.txt")
            .read_text()
            .replace("{{CODE}}", code)
            .replace(
                "{{SOURCE}}",
                str(Path(source).resolve() if isinstance(source, str) else source),
            )
        )
    try:
        if debug:
            code = ast.unparse(ast.parse(code))
            DAHLIA.print(f"&j{code}", file=sys.stderr)
        reg.vars = globals() | reg.vars
        if repl:
            try:
                res = eval(code, reg.vars)
                if not (res is None or res is NULL):
                    print(repr(res))
            except SyntaxError:
                exec(code, reg.vars)
        else:
            exec(code, reg.vars)
    except Exception as e:  # noqa: BLE001
        exc.handle_exception(e)
    Runtime.repl = runtime_state
    return reg



import sys
from re import compile

from dahlia import Dahlia

from samarium.runtime import Runtime

DAHLIA = Dahlia()
NDE_TYPES = {AttributeError, NameError, UnboundLocalError}

ARG_NOT_ITER = compile(r"argument of type '(\w+)' is not iterable")
BAD_OP = compile(
    r"'(.+)' not supported between instances of '(\w+)' and '(\w+)'"
    r"|unsupported operand type\(s\) for (.+): '(\w+)' and '(\w+)'"
)
BAD_UOP = compile(r"bad operand type for unary (.+): '(\w+)'")
NO_GETITEM = compile(r"'(\w+)' object is not subscriptable")
NO_SETITEM = compile(r"'(\w+)' object does not support item assignment")
NOT_CALLITER = compile(r"'(\w+)' object is not (\w+)")
OP_MAP = {
    ">=": ">:",
    "<=": "<:",
    "/": "--",
    "%": "---",
    "*": "++",
    "@": "><",
    "** or pow()": "+++",
}
SINGLE_QUOTED_NAME = compile(r"'(\w+)'")


def clear_name(name: str) -> str:
    return name.removeprefix("__").removeprefix("sm_")


def handle_exception(exception: Exception) -> None:
    exc_type = type(exception)
    errmsg = str(exception)
    name = ""
    if exc_type is NotDefinedError:
        exception = NotDefinedError(".".join(map(clear_name, errmsg.split("."))))
    if (
        exc_type is TypeError
        and "missing 1 required positional argument: 'self'" in errmsg
    ):
        exception = SamariumTypeError("missing instance")
    elif m := BAD_OP.match(errmsg):
        op, lhs, rhs = filter(None, m.groups())
        op = OP_MAP.get(op, op)
        exc_type = NotDefinedError
        exception = exc_type(f"{clear_name(lhs)} {op} {clear_name(rhs)}")
    elif m := BAD_UOP.match(errmsg):
        op, type_ = m.groups()
        exc_type = NotDefinedError
        exception = exc_type(f"{op}{clear_name(type_)}")
    elif m := ARG_NOT_ITER.match(errmsg):
        exc_type = NotDefinedError
        type_ = clear_name(m.group(1))
        exception = exc_type(f"->? {type_}")
    elif m := NO_GETITEM.match(errmsg):
        exc_type = NotDefinedError
        type_ = clear_name(m.group(1))
        exception = exc_type(f"{type_}<<>>")
    elif m := NO_SETITEM.match(errmsg):
        exc_type = NotDefinedError
        type_ = clear_name(m.group(1))
        exception = exc_type(f"{type_}<<>>:")
    elif m := NOT_CALLITER.match(errmsg):
        exc_type = NotDefinedError
        type_ = clear_name(m.group(1))
        template = "... _ ->? {}" if m.group(2) == "iterable" else "{}()"
        exception = exc_type(template.format(type_))
    elif exc_type is SyntaxError:
        exception = SamariumSyntaxError(
            f"invalid syntax at {int(errmsg.split()[-1][:-1])}"
        )
    elif exc_type in NDE_TYPES:
        names = SINGLE_QUOTED_NAME.findall(errmsg)
        exc_type = NotDefinedError
        exception = exc_type(
            f"{clear_name(names[0])}<<>>{':' * (names[1] == '__setitem__')}"
            if len(names) >= 2 and names[1] in {"__getitem__", "__setitem__"}
            else ".".join(map(clear_name, names))
        )
    elif exc_type is ZeroDivisionError:
        name = "MathError"
    elif exc_type not in {AssertionError, NotDefinedError}:
        name = exc_type.__name__.removeprefix("Samarium")
    name = name or exc_type.__name__
    DAHLIA.print(f"&4[{name}] {exception}", file=sys.stderr)
    if not Runtime.repl:
        sys.exit(1)


class SamariumError(Exception):
    pass


class NotDefinedError(SamariumError):
    def __init__(self, inp: object, message: str = "") -> None:
        if isinstance(inp, str):
            message = inp
            inp = ""
        else:
            inp = type(inp).__name__
        super().__init__(f"{inp}.{message}" if inp else message)


class SamariumImportError(SamariumError):
    pass


class SamariumSyntaxError(SamariumError):
    pass


class SamariumTypeError(SamariumError):
    pass


class SamariumValueError(SamariumError):
    pass


class SamariumIOError(SamariumError):
    pass


class SamariumRecursionError(SamariumError):
    pass



from __future__ import annotations

import json
import string
import sys
from ast import literal_eval
from contextlib import redirect_stderr, redirect_stdout, suppress
from dataclasses import dataclass
from datetime import datetime, timedelta
from os import get_terminal_size
from pathlib import Path
from time import time
from typing import TYPE_CHECKING, TypedDict

if TYPE_CHECKING:
    from collections.abc import Iterator

if sys.platform not in ("win32", "cygwin"):
    # used by input() to provide elaborate line editing & history features
    import readline  # noqa: F401

from crossandra import CrossandraError

from samarium import core
from samarium.exceptions import DAHLIA, SamariumSyntaxError, handle_exception
from samarium.runtime import Runtime
from samarium.tokenizer import tokenize
from samarium.transpiler import Registry, match_brackets
from samarium.utils import __version__

TIME = {"t", "time"}
HELP = {"?", "h", "help"}
QUIT = {"!", "exit", "q", "quit", "exit!", "q!", "quit!"}
COMMANDS = (
    {
        "3",
        "clear",
        "color",
        "debug",
        "restore",
        "run",
        "session",
        "undo",
    }
    | HELP
    | TIME
    | QUIT
)
NO_ARG_SESSION_SUBCMDS = {"delete-all", "list", "restore"}

COLOR_TO_CODE = {
    "blue": "&1",
    "green": "&2",
    "cyan": "&3",
    "red": "&4",
    "purple": "&5",
    "orange": "&6",
    "light gray": "&7",
    "gray": "&8",
    "light blue": "&9",
    "black": "&0",
    "lime": "&a",
    "turquoise": "&b",
    "light red": "&c",
    "pink": "&d",
    "yellow": "&e",
    "default": "",
}
CODE_TO_COLOR = {v: k for k, v in COLOR_TO_CODE.items()}

CACHE_DIR = Path.home() / ".cache" / "samarium"
CONFIG_FILE = Path.home() / ".config" / "samarium" / "repl.yaml"
PS1 = "--> "
PS2 = "  > "


class Command:
    def __init__(
        self, *aliases: str, arg: str = "", sep: str = "|", msg: str = ""
    ) -> None:
        self.aliases = aliases
        self.arg = arg
        self.sep = sep
        self.msg = msg

    def __str__(self) -> str:
        return (
            self.sep.join(f"&e{a}&R" for a in self.aliases) + f" &7{self.arg}&R"
        ).ljust(28 + 4 * len(self.aliases)).lstrip() + self.msg


HELP_TEXT = "\n".join(
    map(
        str,
        (
            Command("?", "h", "help", msg="shows this message"),
            Command("exit", "q", "quit", msg="saves the session and quits the repl"),
            Command("!", "exit!", "q!", "quit!", msg="force quits the repl"),
            Command("session", msg="manages sessions, see &2:? session&R for details"),
            Command("clear", msg="clears the screen"),
            Command(
                "color", msg="changes the prompt color, see &2:? color&R for details"
            ),
            Command("debug", msg="toggles debug mode"),
            Command("restore", msg="restores the previous session"),
            Command("t", "time", msg="times the execution of the following statement"),
            Command("undo", msg="undoes the last statement"),
        ),
    )
)

COLOR_HELP_TEXT = """
&ecolor &7[color]&R
&oproviding no color will reset it to the default one&R

use &e:color save&R to save the current color to your config

&00&R|&0black&R
&11&R|&1blue&R
&22&R|&2green&R
&33&R|&3cyan&R
&44&R|&4red&R
&55&R|&5purple&R
&66&R|&6orange&R
&77&R|&7light gray&R
&88&R|&8gray&R
&99&R|&9light blue&R
&aa&R|&alime&R
&bb&R|&bturquoise&R
&cc&R|&clight red&R
&dd&R|&dpink&R
&ee&R|&eyellow&R
"""

SESSIONS_HELP_TEXT = "\n".join(
    map(
        str,
        (
            Command("session autosave", msg="toggles session autosave on exit"),
            Command("session delete-all", msg="deletes all saved sessions"),
            Command(
                "session lifetime",
                arg="[time]",
                msg=(
                    "updates session lifetime (how long they stay cached)\n"
                    f"{' ' * 24}&ocurrent value is shown when run without an argument&R"
                ),
            ),
            Command("session list", msg="lists all sessions and their sizes"),
            Command("session load", arg="<name>", msg="loads a session"),
            Command("session restore", msg="restores the last unnamed session"),
            Command("session save", arg="[name]", msg="saves a session"),
        ),
    )
)

HELP_TEXT_INDEX = {
    "": HELP_TEXT,
    "session": SESSIONS_HELP_TEXT,
    "sessions": SESSIONS_HELP_TEXT,
    "color": COLOR_HELP_TEXT,
}


@dataclass
class REPLConfig:
    color: str = ""
    autosave: bool = True
    session_lifetime: float = 30.0

    @classmethod
    def load(cls) -> REPLConfig:
        lines = CONFIG_FILE.read_text().splitlines()
        data = {
            k.strip(): literal_eval(v.strip())
            for k, _, v in (line.partition(":") for line in lines)
        }
        return cls(**data)

    def save(self) -> None:
        content = "\n".join(f"{k}: {v!r}" for k, v in vars(self).items())
        CONFIG_FILE.write_text(content)


class SessionData(TypedDict):
    color: str
    debug: bool
    history: list[str]


class Session:
    ALLOWED_CHARS = string.ascii_letters + string.digits + "-_"

    def __init__(
        self, color: str, *, debug: bool, history: list[str] | None = None
    ) -> None:
        self.color = color
        self.debug = debug
        self.history = history or []

    @property
    def color(self) -> str:
        return self._color

    @color.setter
    def color(self, value: str) -> None:
        if not value:
            self._color = ""
        elif (code := value.removeprefix("&")) in "0123456789abcde":
            self._color = f"&{code}"
        elif value in COLOR_TO_CODE:
            self._color = COLOR_TO_CODE[value]
        else:
            msg = f"unknown color {value!r}"
            raise ValueError(msg)

    @property
    def color_name(self) -> str:
        return CODE_TO_COLOR.get(self.color, "default")

    @property
    def kwargs(self) -> SessionData:
        return {
            "color": self.color_name,
            "debug": self.debug,
            "history": self.history,
        }

    def copy(self) -> Session:
        kw = self.kwargs
        kw["color"] = self._color
        s = Session(**kw)
        s.history = s.history.copy()
        return s

    def save(self, name: str = "") -> None:
        filename = name or datetime.now().strftime("%Y%m%d%H%M%S")
        data = json.dumps(self.kwargs)
        if not CACHE_DIR.exists():
            CACHE_DIR.mkdir(parents=True, exist_ok=True)
        (CACHE_DIR / f"{filename}.json").write_text(data)

    @classmethod
    def load(cls, name: str) -> Session:
        return cls(**json.loads((CACHE_DIR / f"{name}.json").read_text()))

    @staticmethod
    def is_valid_name(name: str) -> bool:
        return all(char in Session.ALLOWED_CHARS for char in name)


class REPL:
    def __init__(self, *, debug: bool) -> None:
        if not CONFIG_FILE.exists():
            CONFIG_FILE.parent.mkdir(parents=True, exist_ok=True)
            CONFIG_FILE.touch()
        self.config = REPLConfig.load()
        cache_cleanup(self.config.session_lifetime)
        self.registry = Registry(globals())
        self.session = Session(
            COLOR_TO_CODE.get(self.config.color, ""),
            debug=debug,
        )

    def load_session(self, session: Session, *, reset_registry: bool = False) -> None:
        if reset_registry:
            self.registry = Registry(globals())
        with (
            Path("/dev/null").open("w") as dev_null,
            redirect_stdout(dev_null),
            redirect_stderr(dev_null),
        ):
            for stmt in session.history:
                core.run(
                    stmt,
                    self.registry,
                    "",
                    debug=session.debug,
                    load_template=False,
                    repl=True,
                )
        self.session = session

    def finish(self) -> None:
        self.save()
        sys.exit()

    def save(self) -> None:
        if self.config.autosave:
            self.session.save()

    def handle_cmd(self, cmd: str, arg: str) -> str | None:
        if cmd in QUIT:
            if cmd.endswith("!"):
                sys.exit()
            self.finish()
        elif cmd in HELP:
            try:
                DAHLIA.print(HELP_TEXT_INDEX[arg].strip())
            except KeyError:
                repl_err(f"unknown help subsection {arg!r}")
        elif cmd == "color":
            if arg == "save":
                self.config.color = self.session.color_name
                self.config.save()
                return None
            try:
                self.session.color = arg
            except ValueError as e:
                repl_err(str(e))
        elif cmd == "debug":
            self.session.debug = not self.session.debug
        elif cmd in TIME:
            return arg
        elif cmd == "run":
            if not arg:
                repl_err("missing file")
                return None
            try:
                src = Path(arg).read_text()
            except FileNotFoundError:
                repl_err(f"file {arg!r} not found")
                return None
            core.run(
                src, self.registry, "", debug=self.session.debug, load_template=False
            )
        elif cmd == "undo":
            s = self.session.copy()
            s.history.pop()
            self.load_session(s, reset_registry=True)
        elif cmd == "session":
            if not arg:
                repl_err("missing subcommand")
                return None
            try:
                subcmd, arg = arg.split()
            except ValueError:
                subcmd, arg = arg, ""
            if subcmd in NO_ARG_SESSION_SUBCMDS and arg:
                repl_err(f"expected no arguments for subcommand {subcmd!r}")
                return None
            if subcmd == "delete-all":
                sessions = tuple(CACHE_DIR.glob("*"))
                if not sessions:
                    repl_err("no sessions to delete")
                    return None
                if (
                    DAHLIA.input(
                        "Are you sure you want to delete all"
                        f" {len(sessions)} sessions? &7(Y/n) "
                    ).casefold()
                    == "n"
                ):
                    return None
                size = 0
                for session in sessions:
                    size += session.stat().st_size
                    session.unlink()
                DAHLIA.print(f"&aRemoved {fmt_size(size)} of session files")
            elif subcmd == "autosave":
                if arg:
                    if (arg := arg.casefold()) not in ("true", "false"):
                        repl_err(f"invalid value {arg!r} (must be true/false)")
                        return None
                    self.config.autosave = arg == "true"
                    self.config.save()
                if self.config.autosave:
                    DAHLIA.print("&aAutosave enabled")
                else:
                    DAHLIA.print("&cAutosave disabled")
            elif subcmd == "list":
                total_size = 0
                term_size = min(get_terminal_size().columns - 8, 48)
                for session in CACHE_DIR.glob("*"):
                    size = session.stat().st_size
                    total_size += size
                    DAHLIA.print(
                        f"{truncate(session.name, term_size)} &e({fmt_size(size)})"
                    )
                DAHLIA.print("-" * term_size)
                DAHLIA.print(f"&l{'Total':<{term_size + 1}} &e{fmt_size(total_size)}\n")
            elif subcmd == "save":
                if not Session.is_valid_name(arg):
                    repl_err(
                        f"invalid session name {arg!r} "
                        "(allowed chars: A..Z a..z 0..9 _ -])"
                    )
                    return None
                if (CACHE_DIR / f"{arg}.json").exists() and input(
                    f"Session {arg!r} already exists. Replace? (y/N) "
                ).casefold() != "y":
                    return None
                self.session.save(arg)
                DAHLIA.print(f"Saved session {arg!r}")
            elif subcmd == "load":
                if not Session.is_valid_name(arg):
                    repl_err(
                        f"invalid session name {arg!r} "
                        "(allowed chars: A..Z a..z 0..9 _ -])"
                    )
                    return None
                if not (CACHE_DIR / f"{arg}.json").exists():
                    repl_err(f"session {arg!r} not found")
                    return None
                self.load_session(Session.load(arg))
            elif subcmd == "lifetime":
                if not arg:
                    lifetime = str(self.config.session_lifetime).removesuffix(".0")
                    DAHLIA.print(f"Current session lifetime is {lifetime} days")
                    return None
                try:
                    self.config.session_lifetime = float(arg)
                except ValueError:
                    repl_err(f"invalid lifetime {arg!r} (must be a float)")
                    return None
                self.config.save()
                DAHLIA.print(f"&aUpdated session lifetime to {arg} days")
            elif subcmd == "restore":
                try:
                    most_recent_session = max(unnamed_sessions(), key=lambda p: p[1])
                except ValueError:
                    repl_err("no sessions to restore")
                    return None
                self.load_session(Session.load(most_recent_session[0].stem))
                DAHLIA.print(f"&aRestored latest session ({most_recent_session[1]})")
            else:
                repl_err(f"unknown session subcommand {subcmd!r}")
        elif cmd == "clear":
            print("\033[2J\033[H", end="")
        else:  # :3
            DAHLIA.print("&dnya~")
        return None

    def read(self) -> str:
        stmt = ""
        prompt = PS1
        while True:
            stmt += DAHLIA.input(self.session.color + prompt)
            if not stmt:
                continue
            stmt += ";"
            try:
                error, stack = match_brackets(tokenize(stmt, repl=True))
            except CrossandraError as e:
                if '"' in str(e):
                    error, stack = 1, []
                else:
                    handle_exception(SamariumSyntaxError(str(e)))
                    raise RuntimeError from None

            if not error:
                return stmt
            if error == -1:
                handle_exception(
                    SamariumSyntaxError(
                        f"missing opening bracket for {stack[-1].value}"
                    )
                )
                raise RuntimeError from None

            prompt = PS2
            stmt = stmt[:-1] + "\n"

    def run(self) -> None:
        DAHLIA.print(
            f"{self.session.color}Samarium {__version__}"
            + " [DEBUG]" * self.session.debug
        )
        Runtime.repl = True
        while True:
            try:
                time_code = None
                stmt = self.read()
                if is_command(stmt):
                    cmd = stmt[1:-1]
                    if not cmd:
                        repl_err("missing command")
                        continue
                    cmd, _, arg = cmd.partition(" ")
                    if cmd not in COMMANDS:
                        repl_err(f"unknown command {cmd!r}")
                        continue
                    time_code = self.handle_cmd(cmd, arg)
                    if time_code is None:
                        continue
                    time_code += ";"
                elif not stmt.strip("; ").isidentifier():
                    self.session.history.append(stmt)
                start = time()
                core.run(
                    time_code or stmt,
                    self.registry,
                    "",
                    debug=self.session.debug,
                    load_template=False,
                    repl=True,
                )
                if time_code:
                    DAHLIA.print(f"&2{time() - start:.3f} seconds&R")
                self.registry.output = ""
            except RuntimeError:
                pass
            except KeyboardInterrupt:
                print()
            except (EOFError, SystemExit):
                break
        self.finish()


def cache_cleanup(lifetime: float) -> None:
    threshold = datetime.now() - timedelta(days=lifetime)
    for file, date in unnamed_sessions():
        if date < threshold:
            file.unlink()


def fmt_size(size: float) -> str:
    units = iter("KMGTP")
    u = ""
    while size > 1000:
        size /= 1000
        u = next(units)
    return f"{size:.1f}{u}B"


def is_command(stmt: str) -> bool:
    if len(stmt) > 1:
        return stmt[0] == ":" != stmt[1]
    return True


def repl_err(msg: str) -> None:
    DAHLIA.print(f"&4[REPLError] {msg}", file=sys.stderr)


def truncate(string: str, length: int) -> str:
    if len(string) > length:
        return string[: length - 3] + "..."
    return string.ljust(length)


def unnamed_sessions() -> Iterator[tuple[Path, datetime]]:
    for file in CACHE_DIR.glob("*"):
        with suppress(ValueError):
            yield (file, datetime.strptime(file.stem, "%Y%m%d%H%M%S"))



import sys
from contextlib import nullcontext, suppress
from pathlib import Path

from samarium.core import run
from samarium.exceptions import DAHLIA
from samarium.repl import REPL, Command
from samarium.transpiler import Registry
from samarium.utils import __version__

OPTIONS = ("-v", "--version", "-c", "--command", "-h", "--help")

HELP = """samarium &7[option] [-c cmd | file]&R
options and arguments:\n""" + "\n".join(
    map(
        str,
        (
            Command(
                "-c",
                "--command",
                arg="<cmd>",
                sep=", ",
                msg="reads program from string",
            ),
            Command("-h", "--help", sep=", ", msg="shows this message"),
            Command("-v", "--version", sep=", ", msg="prints Samarium version"),
            Command(arg="file", sep=", ", msg=" reads program from script file"),
        ),
    )
)


def main(*, debug: bool = False) -> None:
    reg = Registry(globals())

    if len(sys.argv) == 1:
        return REPL(debug=debug).run()

    if (arg := sys.argv[1]) in OPTIONS:
        if arg in OPTIONS[:2]:
            print(f"Samarium {__version__}")
        elif arg in OPTIONS[2:4]:
            if len(sys.argv) > 2:
                run(sys.argv[2] + " !", reg, arg, debug=debug)
            DAHLIA.print("&4missing code to execute", file=sys.stderr)
        elif arg in OPTIONS[4:]:
            DAHLIA.print(HELP)
        sys.exit()

    try:
        file = Path(arg).read_text()
    except OSError:
        DAHLIA.print(f"&4file not found: {arg}", file=sys.stderr)
    else:
        with nullcontext() if debug else suppress(Exception, KeyboardInterrupt):
            file = "\n".join(file.splitlines()[file.startswith("#!") :])
            run(file, reg, arg, debug=debug)


def main_debug() -> None:
    main(debug=True)



from __future__ import annotations

from re import sub
from typing import TYPE_CHECKING, Any, ClassVar, Generic, TypeVar

from samarium.exceptions import SamariumTypeError, SamariumValueError

if TYPE_CHECKING:
    from collections.abc import Callable

__version__ = "0.6.2"

T = TypeVar("T")


KT = TypeVar("KT")
VT = TypeVar("VT")


class ClassProperty(Generic[T]):
    def __init__(self, func: Callable[..., T]) -> None:
        self.func = func

    def __get__(self, obj: object, owner: object = None) -> T:
        if obj is None:
            obj = owner
        return self.func(obj)


class Singleton:
    _instances: ClassVar[dict[type[Singleton], Singleton]] = {}

    def __new__(cls, *args: Any, **kwargs: Any) -> Singleton:
        if cls not in cls._instances:
            cls._instances[cls] = super().__new__(cls, *args, **kwargs)
        return cls._instances[cls]


def sysexit(*args: Any) -> None:
    if len(args) > 1:
        msg = "=>! only takes one argument"
        raise SamariumTypeError(msg)
    code = args[0].val if args else 0
    if not isinstance(code, (int, str)):
        msg = "=>! only accepts integers and strings"
        raise SamariumTypeError(msg)
    raise SystemExit(code)


def convert_float(string: str, *, base: int, sep: str = ".") -> int | float:
    exp_sep = "p" if base == 16 else "e"
    if not (sep in string or exp_sep in string):
        return int(string, base)
    float_part, _, exp_part = string.partition(exp_sep)
    int_, _, dec = float_part.partition(sep)
    out = int(int_ or "0", base) + sum(
        int(v, base) * base**~i for i, v in enumerate(dec)
    )
    if exp_part:
        try:
            exponent = int(exp_part, 10 if base == 16 else base)
        except ValueError:
            msg = f"invalid exponent: {exp_part}"
            raise SamariumValueError(msg) from None
        out *= (2 if base == 16 else base) ** exponent
    return out


def parse_number(string: str) -> tuple[int | float, bool]:
    string = string.strip()
    orig = string
    neg = len(string)
    b = "d"
    if ":" in string:
        b, _, string = string.partition(":")
        if b not in "box":
            msg = f"{b} is not a valid base"
            raise SamariumValueError(msg)
    base = {"b": 2, "o": 8, "x": 16, "d": 10}[b]
    string = string.lstrip("-")
    neg = -2 * ((neg - len(string)) % 2) + 1

    try:
        num = neg * convert_float(string, base=base)
    except ValueError:
        no_prefix = orig[2:] if orig[1] == ":" else orig
        msg = f'invalid string for Number with base {base}: "{no_prefix}"'
        raise SamariumValueError(msg) from None
    else:
        return num, isinstance(num, int) or num.is_integer()


def smformat(string: str, fields: str | list[Any] | dict[Any, Any]) -> str:
    if isinstance(fields, str):
        fields = [fields]
    it = fields.items() if isinstance(fields, dict) else enumerate(fields)
    for k, v in it:
        string = sub(rf"(?<!\$)\${k}", str(v), string)
    return string.replace("$$", "$")


def get_name(obj: Callable[..., Any] | type) -> str:
    return obj.__name__.removeprefix("sm_")


def get_type_name(obj: Any) -> str:
    return type(obj).__name__.removeprefix("sm_")



from __future__ import annotations

from enum import Enum
from os import write
from pathlib import Path
from typing import IO, TYPE_CHECKING, Any, cast

from samarium.classes.base import NULL, Array, Attrs, Null, Num, Number, Slice, String
from samarium.exceptions import SamariumIOError, SamariumTypeError, SamariumValueError
from samarium.utils import get_type_name

if TYPE_CHECKING:
    from collections.abc import Iterator


class Mode(Enum):
    READ = "r"
    WRITE = "w"
    READ_WRITE = "r+"
    APPEND = "a"


class FileManager:
    @staticmethod
    def create(path: String) -> Null:
        Path(path.val).touch()
        return NULL

    @staticmethod
    def open(path: String | Number, mode: Mode, *, binary: bool = False) -> File:
        if isinstance(path, Number):
            if mode is Mode.READ_WRITE:
                msg = "cannot open a standard stream in a read & write mode"
                raise SamariumIOError(msg)
            if not path.is_int:
                msg = "cannot use non-integers"
                raise SamariumValueError(msg)
        pth = cast("str | int", path.val)
        if isinstance(pth, str):
            f = Path(pth).open(mode.value + "b" * binary)  # noqa: SIM115
        else:
            f = open(pth, mode.value + "b" * binary)  # noqa: PTH123, SIM115
        return File(f, mode.name, pth, binary=binary)

    @staticmethod
    def open_binary(path: String, mode: Mode) -> File:
        return FileManager.open(path, mode, binary=True)

    @staticmethod
    def quick(
        path: String | File | Number,
        mode: Mode,
        *,
        data: String | Array | None = None,
        binary: bool = False,
    ) -> String | Array | Null:
        if isinstance(path, String):
            p = Path(path.val)
            if mode is Mode.READ:
                content = p.read_bytes() if binary else p.read_text()
                return Array(map(Num, content)) if binary else String(content)
            if data is None:
                msg = "missing data"
                raise SamariumIOError(msg)
            if not isinstance(data, Array):
                p.write_text(data.val)
                return NULL
            if isinstance(data, Array):
                bytes_ = b""
                for i in data.val:
                    try:
                        bytes_ += i.val.to_bytes(1, "big")
                    except AssertionError:  # TODO(trag1c): why AssertionError?
                        msg = "some items in the array are not of type Integer"
                        raise SamariumTypeError(msg) from None
                p.write_bytes(bytes_)
        elif isinstance(path, Number):
            if not path.is_int:
                msg = "cannot use non-integers"
                raise SamariumValueError(msg)
            if mode in {Mode.APPEND, Mode.WRITE}:
                fd = cast(int, path.val)
                write(fd, str(data).encode())
            else:
                msg = (
                    "reading from file descriptors is "
                    "not supported for quick operations"
                )
                raise SamariumIOError(msg)
        else:
            file = path
            if mode is Mode.READ:
                return file.load()
            if data is not None:
                file.save(data)
            else:
                msg = "missing data"
                raise SamariumIOError(msg)
        return NULL


class File(Attrs):
    __slots__ = ("binary", "mode", "path", "val")

    def __init__(self, file: IO, mode: str, path: str | int, *, binary: bool) -> None:
        self.binary = binary
        self.mode = mode
        self.path = path
        self.val = file

    def __bool__(self) -> bool:
        return not self.val.closed

    def __str__(self) -> str:
        return f'File(path:"{self.path}", mode:{self.mode})'

    def __invert__(self) -> Null:
        self.val.close()
        return NULL

    def __iter__(self) -> Iterator[String] | Iterator[Array[Number]]:
        if self.binary:
            for line in self.val:
                yield Array(map(Num, list(line)))
        else:
            yield from map(String, self.val)

    def __getitem__(self, index: Any) -> Array | String | Number | Null:
        if isinstance(index, Slice):
            if index.is_empty():
                return Num(self.val.tell())
            if isinstance(index.step, Number):
                msg = "cannot use step"
                raise SamariumIOError(msg)
            if isinstance(index.start, Number):
                if not isinstance(index.stop, Number):
                    return self.load(index.start)
                if not (index.start.is_int or index.stop.is_int):
                    msg = f"invalid index: {index}"
                    raise SamariumValueError(msg)
                start = cast(int, index.start.val)
                stop = cast(int, index.stop.val)
                current_pos = self.val.tell()
                self.val.seek(start)
                data = self.val.read(stop - start)
                self.val.seek(current_pos)
                if self.binary:
                    if isinstance(data, bytes):
                        data = [*data]
                    return Array(map(Num, data))
                return String(data)
            return self[Slice(Num(0), slice.stop, slice.step)]

        self.val.seek(index.val)
        return NULL

    def load(self, bytes_: Number | None = None) -> String | Array:
        if bytes_ is None:
            bytes_ = Num(-1)
        if not bytes_.is_int:
            msg = "cannot use non-integers"
            raise SamariumValueError(msg)
        val = self.val.read(cast(int, bytes_.val))
        if self.binary:
            return Array(map(Num, val))
        return String(val)

    def save(self, data: String | Array) -> Null:
        if (self.binary and isinstance(data, String)) or (
            not self.binary and isinstance(data, Array)
        ):
            raise SamariumTypeError(get_type_name(data))
        if isinstance(data, Array):
            bytes_ = b""
            for i in data.val:
                try:
                    bytes_ += i.val.to_bytes(1, "big")
                except AssertionError:
                    msg = "some items in the array are not of type Integer"
                    raise SamariumTypeError(msg) from None
            self.val.write(bytes_)
        else:
            self.val.write(data.val)
        return NULL
